#
# MIT License
#
# (C) Copyright 2024 Hewlett Packard Enterprise Development LP
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
"""Error handling exceptions used within vTDS and handled by the
wrapper code around main()

"""
import sys


# pylint: disable=too-few-public-methods
class ContextualError(Exception):
    """Exception to report failures seen and contextualized within the
    application providing output and error log file pathnames for more
    detailed error reporting.

    """
    def __init__(self, msg, output=None, error=None):
        """Constructor.

        """
        self.output = output
        self.error = error
        super().__init__(msg)

    def __str__(self):
        """String conversion

        """
        result = super().__str__()
        files = isinstance(self.output, str) or isinstance(self.error, str)
        result += " [" if files else ""
        result += (
            "standard output log in: '%s'" % self.output
        ) if isinstance(self.output, str) else ""
        result += (
            ", "
        ) if files else ""
        result += (
            "standard error log in: '%s'" % self.error
        ) if isinstance(self.error, str) else ""
        result += (
            "]"
        ) if files else ""
        return result


class UsageError(Exception):  # pylint: disable=too-few-public-methods
    """Exception to report usage errors

    """


def write_out(string):
    """Write an arbitrary string on stdout and make sure it is
    flushed.

    """
    sys.stdout.write(string)
    sys.stdout.flush()


def write_err(string):
    """Write an arbitrary string on stderr and make sure it is
    flushed.

    """
    sys.stderr.write(string)
    sys.stderr.flush()


def usage(usage_msg, err=None):
    """Print a usage message and exit with an error status.

    """
    if err:
        write_err("ERROR: %s\n" % err)
    write_err("%s\n" % usage_msg)
    sys.exit(1)


def error_msg(msg):
    """Format an error message and print it to stderr.

    """
    write_err("ERROR: %s\n" % msg)


def warning_msg(msg):
    """Format a warning and print it to stderr.

    """
    write_err("WARNING: %s\n" % msg)


def info_msg(msg):
    """Format an informational message and print it to stderr.

    """
    write_err("INFO: %s\n" % msg)



#
# MIT License
#
# (C) Copyright 2024 Hewlett Packard Enterprise Development LP
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
"""Init script for the vTDS base library

"""
from .entrypoint import (
    entrypoint
)
from .errors import (
    ContextualError,
    UsageError,
    write_out,
    write_err,
    error_msg,
    warning_msg,
    info_msg
)
from .config_operations import (
    read_config,
    merge_configs,
    expand_inheritance
)

from .base_config import (
    BaseConfiguration
)

from .template_operations import (
    render_template_file,
    render_templated_tree,
    render_command_string
)

from .stack import (
    VTDSStack
)
from .logs import (
    log_paths,
    logfile
)
from .commands import (
    run
)



#
# MIT License
#
# (C) Copyright [2024] Hewlett Packard Enterprise Development LP
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
"""Public API for the GCP provider layer base configuration data, this
gives callers access to the Provider's BaseConfig API and prevents
them from seeing the private GCP specific implementation of the API.

"""
from abc import (
    ABCMeta,
    abstractmethod
)


class BaseConfig(meta_class=ABCMeta):
    """BaseConfig class presents operations on the base configuration
    of the provider layer to callers.

    """
    @abstractmethod
    def get_base_config(self):
        """Retrieve the base configuration for the provider in the
        form of a python data structure for use in composing and
        overall vTDS configuration.

        """

    @abstractmethod
    def get_base_config_text(self):
        """Retrieve the text of the base configuration file as a text
        string (UTF-8 encoded) for use in displaying the configuration
        to users.

        """

    @abstractmethod
    def get_test_overlay(self):
        """Retrieve a pre-defined test overlay configuration in the
        form of a python data structure for use in composing vTDS
        configurations for testing with this provider layer.

        """



#
# MIT License
#
# (C) Copyright 2024 Hewlett Packard Enterprise Development LP
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
"""Operations on configuration structures.

"""

from yaml import (
    YAMLError,
    safe_load
)
from .errors import ContextualError


def merge_configs(base, overlay):
    """Given a base configuration specified in 'base', merge an
    overlay configuration on top of that base configuration to form a
    new configuration that is the merged configuration. Data in the
    overlay override the contents of the base as follows:

    - Key / value pairs that exist in both and have scalar, list or
      class values are completely taken from the overlay (base
      content is discarded).

    - Key / value pairs that exist in both and have dictionaries are
      recursively merged using these rules.

    - Key / value pairs found only in the base are taken from the base

    - Key value pairs found only in the overlay are taken from the
      overlay.

    """
    # If either the base or the overlay is not a dictionary, then the
    # value going into the config at this layer (or the whole config
    # if we are in the zero-th recursion) is simply the overlay.
    if not isinstance(base, dict) or not isinstance(overlay, dict):
        return overlay

    # Both are dictionaries, so we are going to merge them. Make a new
    # dictionary to hold the merged configs at this level.
    new_config = {}

    # Populate the new config with all of the items that are in the
    # base appropriately overlaid.
    for key, value in base.items():
        if isinstance(value, dict):
            if key in overlay:
                # Base has a dictionary and overlay has something,
                # merge the two (if the overlay isn't a dictionary,
                # the check at the beginning kicks in and returns the
                # overlay value).
                new_config[key] = merge_configs(base[key], overlay[key])
                continue
            # Base has a dictionary, but the overlay has nothing, keep
            # the base.
            new_config[key] = value
            continue
        # Not a dictionary, overwrite from the overlay?
        if key in overlay:
            # Yep! Take the overlay value
            new_config[key] = overlay[key]
            continue
        # Nope. Keep the base value
        new_config[key] = value

    # Now, go through the overlay and catch any keys that aren't in
    # the base. If they were in both the base and overlay, we handled
    # then above, so we only care about what is only in the overlay,
    # and we just want to add it as is to the new config.
    for key, value in overlay.items():
        if key not in base:
            new_config[key] = value

    # new_config now contains the merged dictionary, return it.
    return new_config


def expand_inheritance(configs, config_name, ancestry=None, top_config=None):
    """Within a set of configurations contained in 'configs', perform an
    inheritace expansion of the configuration named 'config_name' that produces
    a fully populated configuation taking into account all of the
    ancestor configurations of 'config_name', and return the data structure
    containing the expanded configuration.

    The 'ancestry' and 'top_config' arguments are used internally during
    recursion and should not be set by the external caller.

    """
    # Set up an 'ancestry' array consisting of the config names we
    # have visited in the inheritance chain so we can tell if we hit
    # an inheritance loop.
    if ancestry is None:
        ancestry = []
    ancestry.append(config_name)
    if top_config is None:
        top_config = config_name
    try:
        config = configs[config_name]
    except KeyError as err:
        raise ContextualError(
            "cannot find config sub-tree named '%s' in config list %s" % (
                config_name, str(configs.keys())
            )
        ) from err
    if 'pure_base_class' not in config:
        # Make sure 'config' does not inherit the 'pure_base_class'
        # designation from any of its parent classes.
        config['pure_base_class'] = False
    parent = config.get('parent_class', None)
    if parent is None:
        # If this config does not inherit from anywhere, then we are
        # done, just return the content of the config.
        return config
    if parent in ancestry:
        # We have already seen 'parent' in the inheritiance chain,
        # this is a loop, so raise an exception.
        raise ContextualError(
            "the inheritance chain for '%s' has a circular dependency "
            "on '%s' - %s" % (
                top_config, parent, str(ancestry + [parent])
            )
        )
    parent_config = expand_inheritance(configs, parent, ancestry, top_config)
    # Merge the current config on top of the parent(s) that have
    # already been expanded, then set 'parent_class' to None because
    # this config is now fully expanded and does not need (or want)
    # further expansion due to inheritance.
    config = merge_configs(parent_config, config)
    config['parent_class'] = None
    return config


def read_config(path, description="configuration"):
    """Read in a YAML configuration from a file.

    """
    try:
        with open(path, 'r', encoding='UTF-8') as config_stream:
            return safe_load(config_stream)
    except OSError as err:
        raise ContextualError(
            "cannot open %s '%s' - %s" % (description, path, str(err))
        ) from err
    except YAMLError as err:
        raise ContextualError(
            "error parsing %s "
            "'%s' - %s" % (description, path, str(err))
        ) from err



#
# MIT License
#
# (C) Copyright 2025 Hewlett Packard Enterprise Development LP
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR
# OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE,
# ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
# OTHER DEALINGS IN THE SOFTWARE.
"""Operations on Jinja templated files and trees filled with Jinja
templated files.

"""

from glob import glob
from jinja2 import (
    Template,
    TemplateError
)

from .errors import ContextualError


def render_template_file(path, data, outpath=None):
    """Render the Jinja template found in 'path' using the parameters
        found in the dictionary supplied by 'data. The file supplied
        in 'path' is overwritten with the rendered template and not
        preserved anywhere.

    """
    outpath = path if outpath is None else outpath
    try:
        with open(path, 'r', encoding="UTF-8") as template_file:
            template_data = template_file.read()
            template = Template(template_data)
            rendered = template.render(data)
    except OSError as err:
        raise ContextualError(
            "cannot read Jinja template file %s: %s" % (
                path, str(err)
            )
        ) from err
    except TemplateError as err:
        raise ContextualError(
            "error rendering template file '%s' - %s" %
            (path, str(err))
        ) from err
    try:
        with open(outpath, 'w', encoding="UTF-8") as output_file:
            output_file.write(rendered)
    except OSError as err:
        raise ContextualError(
            "cannot write Jinja output file %s: %s" % (
                path, str(err)
            )
        ) from err


def render_templated_tree(patterns, data, build_dir):
    """Render the all of the files matching a pattern in 'patterns'
    found in the directory 'build_dir' in place as Jinja templates
    using the template data provided in 'data'.

    """
    # For each file name pattern specified, get the list of files
    # found in the build tree matching that pattern and render each
    # one.
    for pattern in patterns:
        paths = glob("%s/**/%s" % (build_dir, pattern), recursive=True)
        for path in paths:
            render_template_file(path, data)


def render_command_string(cmd, jinja_values):
    """Render a command string (not a command list) as a Jinja
    template with substitutions from the supplied jinja_values
    dictionary.

    """
    try:
        template = Template(cmd)
        return template.render(**jinja_values)
    except TemplateError as err:
        raise ContextualError(
            "error using Jinja to render command line '%s' - %s" % (
                cmd,
                str(err)
            )
        ) from err
